'''The standard Sphysics dam break benchmark. (4 hours)

This is intended to be the same as the standard dam break from the
DualSPhysics code just for comparison.
'''
import numpy as np

from pysph.base.utils import get_particle_array
from pysph.solver.application import Application
from pysph.sph.scheme import WCSPHScheme


def ravel(*args):
    return tuple(np.ravel(x) for x in args)


def rhstack(*args):
    '''Join given set of args, we that each element in args has the same shape.
    Each argument is first ravelled and then stacked.
    '''
    return tuple(np.hstack(ravel(*t)) for t in zip(*args))


class DamBreak(Application):
    def add_user_options(self, group):
        group.add_argument(
            '--dx', action='store', type=float, dest='dx',  default=0.0085,
            help='Particle spacing.'
        )
        hdx = np.sqrt(3)
        group.add_argument(
            '--hdx', action='store', type=float, dest='hdx',  default=hdx,
            help='Specify the hdx factor where h = hdx * dx.'
        )

    def consume_user_options(self):
        dx = self.options.dx
        self.dx = dx
        self.hdx = self.options.hdx

    def create_scheme(self):
        self.c0 = c0 = 10.0 * np.sqrt(2.0*9.81*0.3)
        self.hdx = hdx = np.sqrt(3)
        dx = 0.01
        h0 = hdx*dx
        alpha = 0.1
        beta = 0.0
        gamma = 7.0
        s = WCSPHScheme(
            ['fluid'], ['boundary'], dim=3, rho0=1000, c0=c0,
            h0=h0, hdx=hdx, gz=-9.81, alpha=alpha, beta=beta, gamma=gamma,
            hg_correction=True, tensile_correction=False
        )
        return s

    def configure_scheme(self):
        s = self.scheme
        h0 = self.dx * self.hdx
        s.configure(h0=h0, hdx=self.hdx)
        dt = 0.25*h0/(1.1 * self.c0)
        tf = 1.5
        s.configure_solver(
            tf=tf, dt=dt, adaptive_timestep=True, n_damp=50
        )

    def create_particles(self):
        dx = self.dx
        l, b, h = 1.6, 0.67, 0.4
        lw, hw = 0.4, 0.3

        # Big filled vessel with staggered points.
        x, y, z = np.mgrid[0:l+dx:dx, 0:b+dx:dx, 0:h:dx]

        # The post
        x3, y3, z3 = np.mgrid[0.9:1.02:dx, 0.25:0.37:dx, dx:0.45:dx]
        xmax, ymax, zmax = max(x3.flat), max(y3.flat), max(z3.flat)
        post_cond = ~((x3 > 0.9) & (x3 < xmax) & (y3 > 0.25) &
                      (y3 < ymax) & (z3 < zmax))
        p_post = x3[post_cond], y3[post_cond], z3[post_cond]

        # Masks to extract different parts from the vessel.
        wcond = ((x > 0) & (x < lw) & (y > 0) &
                 (y < b) & (z > 0) & (z < hw))
        box = ~((x > 0) & (x <= l) & (y > 0) & (y < b) & (z > 0) & (z <= h))

        p_box = x[box], y[box], z[box]
        xf, yf, zf = x[wcond], y[wcond], z[wcond]

        xs, ys, zs = rhstack(p_box, p_post)

        vol = dx**3
        m = vol*1000
        f = get_particle_array(
            name='fluid', x=xf, y=yf, z=zf, m=m, h=dx*self.hdx,
            rho=1000.0
        )
        b = get_particle_array(
            name='boundary', x=xs, y=ys, z=zs, m=m, h=dx*self.hdx,
            rho=1000.0
        )

        self.scheme.setup_properties([f, b])
        return [f, b]

    def customize_output(self):
        self._mayavi_config('''
        viewer.scalar = 'vmag'
        b = particle_arrays['boundary']
        b.plot.actor.mapper.scalar_visibility = False
        b.plot.actor.property.opacity = 0.15
        ''')


if __name__ == '__main__':
    app = DamBreak()
    app.run()



# cython: language_level=3, embedsignature=True
# distutils: language=c++
from libcpp.vector cimport vector

from .nnps_base cimport *

#Imports for SpatialHashNNPS
cdef extern from "spatial_hash.h":
    cdef cppclass HashEntry:
        double h_max

        vector[unsigned int] *get_indices() nogil

    cdef cppclass HashTable:
        HashTable(long long int) nogil except +
        void add(int, int, int, int, double) nogil
        HashEntry* get(int, int, int) nogil

# NNPS using Spatial Hashing algorithm
cdef class SpatialHashNNPS(NNPS):
    ############################################################################
    # Data Attributes
    ############################################################################
    cdef long long int table_size               # Size of hashtable
    cdef double radius_scale2

    cdef HashTable** hashtable
    cdef HashTable* current_hash

    cdef NNPSParticleArrayWrapper dst, src

    ##########################################################################
    # Member functions
    ##########################################################################

    cpdef set_context(self, int src_index, int dst_index)

    cdef void find_nearest_neighbors(self, size_t d_idx, UIntArray nbrs) nogil

    cdef inline void _add_to_hashtable(self, int hash_id, unsigned int pid, double h,
            int i, int j, int k) nogil

    cdef inline int _neighbor_boxes(self, int i, int j, int k,
            int* x, int* y, int* z) nogil

    cpdef _refresh(self)

    cpdef _bin(self, int pa_index, UIntArray indices)

# NNPS using Extended Spatial Hashing algorithm
cdef class ExtendedSpatialHashNNPS(NNPS):
    ############################################################################
    # Data Attributes
    ############################################################################
    cdef long long int table_size               # Size of hashtable
    cdef double radius_scale2

    cdef HashTable** hashtable
    cdef HashTable* current_hash

    cdef NNPSParticleArrayWrapper dst, src

    cdef int H
    cdef double h_sub
    cdef bint approximate

    ##########################################################################
    # Member functions
    ##########################################################################

    cpdef set_context(self, int src_index, int dst_index)

    cdef void find_nearest_neighbors(self, size_t d_idx, UIntArray nbrs) nogil

    cdef inline int _h_mask_approx(self, int* x, int* y, int* z) nogil

    cdef inline int _h_mask_exact(self, int* x, int* y, int* z) nogil

    cdef int _neighbor_boxes(self, int i, int j, int k,
            int* x, int* y, int* z, double h) nogil

    cdef inline void _add_to_hashtable(self, int hash_id, unsigned int pid, double h,
            int i, int j, int k) nogil

    cpdef _refresh(self)

    cpdef _bin(self, int pa_index, UIntArray indices)



"""
Inlet Outlet Manager
"""

from pysph.sph.equation import Equation
from compyle.api import get_config
from pysph.base.utils import get_particle_array
from pysph.sph.integrator_step import IntegratorStep
import numpy as np
from compyle.api import declare


class InletInfo(object):
    def __init__(self, pa_name, normal, refpoint, has_ghost=True,
                 update_cls=None, equations=None, umax=1.0,
                 props_to_copy=None):
        """Create object with information of inlets, all the others parameters which
           are not passed here get evaluated by `InletOutletManager` once the
           inlet is created.

        Parameters
        ----------

        pa_name : str
            Name of the inlet
        normal : list
            Components of normal (float)
        refpoint : list
            Point at the fluid-inlet interface (float)
        has_ghost : bool
            if True, the ghost particles will be created
        update_cls : class_name
            the class which is to be used to update the inlet/outlet
        equations : list
            List of equations (optional)
        props_to_copy : array
            properties to copy

        """

        self.pa_name = pa_name
        self.normal = normal
        self.refpoint = refpoint
        self.has_ghost = has_ghost
        self.update_cls = InletBase if update_cls is None else update_cls
        self.length = 0.0
        self.dx = 0.1
        self.umax = umax
        self.equations = [] if equations is None else equations
        self.props_to_copy = props_to_copy


class OutletInfo(InletInfo):
    def __init__(self, pa_name, normal, refpoint, has_ghost=False,
                 update_cls=None, equations=None, umax=1.0,
                 props_to_copy=None):
        """Create object with information of outlet

        The name is kept different for distinction only.
        """
        super(OutletInfo, self).__init__(
            pa_name, normal, refpoint, has_ghost, update_cls,
            equations, umax, props_to_copy)
        self.update_cls = OutletBase if update_cls is None else update_cls


class InletOutletManager(object):
    def __init__(self, fluid_arrays, inletinfo, outletinfo,
                 extraeqns=None):
        """Create the object to manage inlet outlet boundary conditions.
           Most of the variables are evaluated after the scheme and particles
           are created.

        Parameters
        -----------

        fluid_arrays : list
            List of fluid particles array names (str)
        inletinfo : list
            List of inlets (InletInfo)
        outletinfo : list
            List of outlet (OutletInfo)
        extraeqns : dict
            Dict of custom equations
        """
        self.fluids = fluid_arrays
        self.dim = None
        self.kernel = None
        self.inlets = [] if inletinfo is None else\
            [x.pa_name for x in inletinfo]
        self.outlets = [] if outletinfo is None else\
            [x.pa_name for x in outletinfo]
        self.inlet_pairs = {}
        self.outlet_pairs = {}
        self.inletinfo = inletinfo
        self.outletinfo = outletinfo
        self.ghost_inlets = []
        self.ghost_outlets = []
        self.inlet_pairs = {}
        self.outlet_pairs = {}
        self.extraeqns = {} if extraeqns is None else extraeqns
        self.active_stages = []
        self._create_ghost_names()

    def create_ghost(self, pa_arr, inlet=True):
        """Creates ghosts for the given inlet/outlet particles

        Parameters
        -----------

        pa_arr : Particle array
            particles array for which ghost is required
        inlet : bool
            if True, inlet info will be used for ghost creation
        """
        xref, yref, zref = 0.0, 0.0, 0.0
        xn, yn, zn = 0.0, 0.0, 0.0
        has_ghost = True
        if inlet:
            for info in self.inletinfo:
                if info.pa_name == pa_arr.name:
                    xref = info.refpoint[0]
                    yref = info.refpoint[1]
                    zref = info.refpoint[2]
                    xn = info.normal[0]
                    yn = info.normal[1]
                    zn = info.normal[2]
                    has_ghost = info.has_ghost
                    break

        if not inlet:
            for info in self.outletinfo:
                if info.pa_name == pa_arr.name:
                    xref = info.refpoint[0]
                    yref = info.refpoint[1]
                    zref = info.refpoint[2]
                    xn = info.normal[0]
                    yn = info.normal[1]
                    zn = info.normal[2]
                    has_ghost = info.has_ghost
                    break

        if not has_ghost:
            return None
        x = pa_arr.x
        y = pa_arr.y
        z = pa_arr.z

        xij = x - xref
        yij = y - yref
        zij = z - zref

        disp = xij * xn + yij * yn + zij * zn
        x = x - 2 * disp * xn
        y = y - 2 * disp * yn
        z = z - 2 * disp * zn

        m = pa_arr.m
        h = pa_arr.h
        rho = pa_arr.rho
        u = pa_arr.u
        name = ''
        if inlet:
            name = self.inlet_pairs[pa_arr.name]
        else:
            name = self.outlet_pairs[pa_arr.name]

        ghost_pa = get_particle_array(
            name=name, m=m, x=x, y=y, h=h, u=u, p=0.0, rho=rho
        )

        return ghost_pa

    def _create_ghost_names(self):
        '''Creates names for ghost for both inlets and outlets if needed'''
        for inlet in self.inletinfo:
            if inlet.has_ghost:
                name = 'ghost_' + inlet.pa_name
                self.inlet_pairs[inlet.pa_name] = name
                self.ghost_inlets.append(name)

        for outlet in self.outletinfo:
            if outlet.has_ghost:
                name = 'ghost_' + outlet.pa_name
                self.outlet_pairs[outlet.pa_name] = name
                self.ghost_outlets.append(name)

    def update_dx(self, dx):
        """Update the discretisation length

        Parameters
        -----------

        dx : float
            The discretisation length
        """
        all_info = self.inletinfo + self.outletinfo
        for info in all_info:
            info.dx = dx

    def _update_inlet_outlet_info(self, pa):
        """Updates refpoint, length and bound

        Parameters
        -----------

        pa : Particle_array
            Particle array of inlet/outlet
        """
        all_info = self.inletinfo + self.outletinfo
        for info in all_info:
            dx = info.dx
            if (info.pa_name == pa.name):
                x = pa.x
                y = pa.y
                z = pa.z
                xmax, xmin = max(x)+dx/2, min(x)-dx/2
                ymax, ymin = max(y)+dx/2, min(y)-dx/2
                zmax, zmin = max(z)+dx/2, min(z)-dx/2
                xdist = xmax - xmin
                ydist = ymax - ymin
                zdist = zmax - zmin
                xn, yn, zn = info.normal[0], info.normal[1], info.normal[2]
                info.length = abs(xdist*xn+ydist*yn+zdist*zn)

    def add_io_properties(self, pa, scheme=None):
        """Add properties to be used in inlet/outlet equations

        Parameters
        ----------

        pa : particle_array
            Particle array of inlet/outlet
        scheme : pysph.sph.scheme
            The insance of scheme class
        """
        pass

    def get_io_names(self, ghost=False):
        """return all the names of inlets and outlets
        Parameters
        ----------

        ghost : bool
            if True, return the names of ghost also
        """
        if ghost:
            return (self.inlets + self.outlets + self.ghost_inlets +
                    self.ghost_outlets)
        else:
            return self.inlets + self.outlets

    def get_stepper(self, scheme, integrator, **kw):
        """Returns the steppers for inlet/outlet

        Parameters
        ----------

        scheme : pysph.sph.scheme
            The instance of the scheme class
        intergrator : pysph.sph.integrator
            The parent class of the integrator
        **kw : extra arguments
            Extra arguments depending upon the scheme used
        """
        raise NotImplementedError()

    def setup_iom(self, dim, kernel):
        """Essential data passed

        Parameters
        ----------

        dim : int
            dimension of the problem
        kernel : pysph.base.kernel
            the kernel instance
        """
        self.dim = dim
        self.kernel = kernel

    def get_equations(self, scheme, **kw):
        """Returns the equations for inlet/outlet

        Parameters
        ----------

        scheme : pysph.sph.scheme
            The instance of the scheme class
        **kw : extra arguments
            Extra arguments depending upon the scheme used
        """
        return []

    def get_equations_post_compute_acceleration(self):
        """Returns the equations for inlet/outlet used post acceleration
           computation"""
        return []

    def get_inlet_outlet(self, particle_array):
        """ Returns list of `Inlet` and `Outlet` instances which
            performs the change in inlet particles to outlet
            particles.

        Parameters
        -----------

        particle_array : list
            List of all particle_arrays
        """
        result = []
        for inlet in self.inletinfo:
            i_name = inlet.pa_name
            self._update_inlet_outlet_info(particle_array[i_name])
            ghost_pa = None if i_name not in self.inlet_pairs\
                else particle_array[self.inlet_pairs[i_name]]
            for fluid in self.fluids:
                in1 = inlet.update_cls(
                        particle_array[i_name],
                        particle_array[fluid],
                        inlet,
                        self.kernel,
                        self.dim,
                        self.active_stages,
                        ghost_pa=ghost_pa
                        )
            result.append(in1)

        for outlet in self.outletinfo:
            o_name = outlet.pa_name
            self._update_inlet_outlet_info(particle_array[o_name])
            ghost_pa = None if o_name not in self.outlet_pairs else\
                particle_array[self.outlet_pairs[o_name]]
            for fluid in self.fluids:
                out1 = outlet.update_cls(
                        particle_array[o_name],
                        particle_array[fluid],
                        outlet,
                        self.kernel,
                        self.dim,
                        self.active_stages,
                        ghost_pa=ghost_pa
                        )
            result.append(out1)

        return result


class IOEvaluate(Equation):
    def __init__(self, dest, sources, x, y, z, xn, yn, zn,
                 maxdist=1000.0):
        """Compute ioid for the particles
           0 : particle is in fluid
           1 : particle is inside the inlet/outlet
           2 : particle is out of inlet/outlet

        Parameters
        -----------

        dest : str
            destination particle array name
        sources : list
            List of source particle arrays
        x : float
            x coordinate of interface point
        y : float
            y coordinate of interface point
        z : float
            z coordinate of interface point
        xn : float
            x component of interface outward normal
        yn : float
            y component of interface outward normal
        zn : float
            z component of interface outward normal
        maxdist : float
            Maximum length of inlet/outlet

        """
        self.x = x
        self.y = y
        self.z = z
        self.xn = xn
        self.yn = yn
        self.zn = zn
        self.maxdist = maxdist

        super(IOEvaluate, self).__init__(dest, sources)

    def initialize(self, d_ioid, d_idx):
        d_ioid[d_idx] = 1

    def loop(self, d_idx, d_x, d_y, d_z, d_ioid, d_disp):
        delx = d_x[d_idx] - self.x
        dely = d_y[d_idx] - self.y
        delz = d_z[d_idx] - self.z

        d_disp[d_idx] = delx * self.xn + dely * self.yn + delz * self.zn

        if ((d_disp[d_idx] > 0.000001) and
           (d_disp[d_idx]-self.maxdist < 0.000001)):
            d_ioid[d_idx] = 1
        elif (d_disp[d_idx] - self.maxdist > 0.000001):
            d_ioid[d_idx] = 2
        else:
            d_ioid[d_idx] = 0


class UpdateNormalsAndDisplacements(Equation):
    def __init__(self, dest, sources, xn, yn, zn, xo, yo, zo):
        """Update normal and perpendicular distance from the interface
        for the inlet/outlet particles

        Parameters
        ----------

        dest : str
            destination particle array name
        sources : list
            List of source particle arrays
        xn : float
            x component of interface outward normal
        yn : float
            y component of interface outward normal
        zn : float
            z component of interface outward normal
        xo : float
            x coordinate of interface point
        yo : float
            y coordinate of interface point
        zo : float
            z coordinate of interface point
        """
        self.xn = xn
        self.yn = yn
        self.zn = zn
        self.xo = xo
        self.yo = yo
        self.zo = zo

        super(UpdateNormalsAndDisplacements, self).__init__(dest, sources)

    def loop(self, d_idx, d_xn, d_yn, d_zn, d_x, d_y, d_z, d_disp):
        d_xn[d_idx] = self.xn
        d_yn[d_idx] = self.yn
        d_zn[d_idx] = self.zn

        xij = declare('matrix(3)')
        xij[0] = d_x[d_idx] - self.xo
        xij[1] = d_y[d_idx] - self.yo
        xij[2] = d_z[d_idx] - self.zo

        d_disp[d_idx] = abs(xij[0]*d_xn[d_idx] + xij[1]*d_yn[d_idx] +
                            xij[2]*d_yn[d_idx])


class CopyNormalsandDistances(Equation):
    """Copy normals and distances from outlet/inlet particles to ghosts"""

    def initialize_pair(self, d_idx, d_xn, d_yn, d_zn, s_xn, s_yn, s_zn,
                        d_disp, s_disp):
        d_xn[d_idx] = s_xn[d_idx]
        d_yn[d_idx] = s_yn[d_idx]
        d_zn[d_idx] = s_zn[d_idx]

        d_disp[d_idx] = s_disp[d_idx]


class InletStep(IntegratorStep):
    def initialize(self, d_x0, d_idx, d_x):
        d_x0[d_idx] = d_x[d_idx]

    def stage1(self, d_idx, d_x, d_x0, d_u, dt):
        dtb2 = 0.5 * dt
        d_x[d_idx] = d_x0[d_idx] + dtb2*d_u[d_idx]

    def stage2(self, d_idx, d_x, d_x0, d_u, dt):
        d_x[d_idx] = d_x0[d_idx] + dt*d_u[d_idx]


class OutletStepWithUhat(IntegratorStep):
    def initialize(self, d_x0, d_idx, d_x):
        d_x0[d_idx] = d_x[d_idx]

    def stage1(self, d_idx, d_x, d_x0, d_uhat, dt):
        dtb2 = 0.5 * dt
        d_x[d_idx] = d_x0[d_idx] + dtb2*d_uhat[d_idx]

    def stage2(self, d_idx, d_x, d_x0, d_uhat, dt):
        d_x[d_idx] = d_x0[d_idx] + dt*d_uhat[d_idx]


class OutletStep(InletStep):
    pass


class InletBase(object):
    def __init__(self, inlet_pa, dest_pa, inletinfo, kernel, dim,
                 active_stages=[1], callback=None, ghost_pa=None):
        """An API to add/delete particle when moving between inlet-fluid

        Parameters
        ----------

        inlet_pa : particle_array
            particle array for inlet
        dest_pa : particle_array
            particle_array of the fluid
        inletinfo : InletInfo instance
            contains information fo inlet
        kernel : Kernel instance
            Kernel to be used for computations
        dim : int
            dimension of the problem
        active_stages : list
            stages of integrator at which update should be active
        callback : function
            callback after the update function
        ghost_pa : particle_array
            particle_array of the ghost_inlet
        """
        self.inlet_pa = inlet_pa
        self.dest_pa = dest_pa
        self.ghost_pa = ghost_pa
        self.callback = callback
        self.dim = dim
        self.kernel = kernel
        self.inletinfo = inletinfo
        self.x = self.y = self.z = 0.0
        self.xn = self.yn = self.zn = 0.0
        self.length = 0.0
        self.dx = 0.0
        self.active_stages = active_stages
        self.io_eval = None
        self._init = False
        cfg = get_config()
        self.gpu = cfg.use_opencl or cfg.use_cuda

    def initialize(self):
        """Function to initialize the class variables after
        evaluation in SimpleInletOutlet class"""
        inletinfo = self.inletinfo
        self.x = inletinfo.refpoint[0]
        self.y = inletinfo.refpoint[1]
        self.z = inletinfo.refpoint[2]
        self.xn = inletinfo.normal[0]
        self.yn = inletinfo.normal[1]
        self.zn = inletinfo.normal[2]
        self.length = inletinfo.length
        self.dx = inletinfo.dx

    def _create_io_eval(self):
        """Evaluator to assign ioid to particles leaving a domain"""
        if self.io_eval is None:
            from pysph.sph.equation import Group
            from pysph.tools.sph_evaluator import SPHEvaluator
            i_name = self.inlet_pa.name
            f_name = self.dest_pa.name
            eqns = []
            eqns.append(Group(equations=[
                IOEvaluate(
                    i_name, [], x=self.x, y=self.y, z=self.z,
                    xn=self.xn, yn=self.yn, zn=self.zn,
                    maxdist=self.length)],
                real=False, update_nnps=False))

            eqns.append(Group(equations=[
                IOEvaluate(
                    f_name, [], x=self.x, y=self.y, z=self.z, xn=self.xn,
                    yn=self.yn, zn=self.zn)],
                real=False, update_nnps=False))

            if self.gpu:
                from pysph.base.gpu_nnps import ZOrderGPUNNPS as NNPS
            else:
                from pysph.base.nnps import LinkedListNNPS as NNPS

            arrays = [self.inlet_pa] + [self.dest_pa]
            io_eval = SPHEvaluator(
                arrays=arrays, equations=eqns, dim=self.dim,
                kernel=self.kernel, nnps_factory=NNPS)
            return io_eval
        else:
            return self.io_eval

    def update(self, time, dt, stage):
        """ Update function called after each stage"""
        if not self._init:
            self.initialize()
            self._init = True
        if stage in self.active_stages:

            dest_pa = self.dest_pa
            inlet_pa = self.inlet_pa
            ghost_pa = self.ghost_pa

            self.io_eval = self._create_io_eval()
            self.io_eval.update()
            self.io_eval.evaluate()

            if self.gpu:
                inlet_pa.gpu.pull(*'ioid x y z'.split())
                ghost_pa.gpu.pull(*'x y z'.split())
                dest_pa.gpu.pull('ioid')
            io_id = inlet_pa.ioid
            cond = (io_id == 0)
            all_idx = np.where(cond)[0]
            inlet_pa.extract_particles(all_idx, dest_pa)

            # moving the moved particles back to the array beginning.
            inlet_pa.x[all_idx] += self.length * self.xn
            inlet_pa.y[all_idx] += self.length * self.yn
            inlet_pa.z[all_idx] += self.length * self.zn

            if ghost_pa:
                ghost_pa.x[all_idx] -= self.length * self.xn
                ghost_pa.y[all_idx] -= self.length * self.yn
                ghost_pa.z[all_idx] -= self.length * self.zn

            if self.callback is not None:
                self.callback(dest_pa, inlet_pa)


class OutletBase(object):
    def __init__(self, outlet_pa, source_pa, outletinfo, kernel,
                 dim, active_stages=[1], callback=None, ghost_pa=None):
        """An API to add/delete particle when moving between fluid-outlet

        Parameters
        ----------

        outlet_pa : particle_array
            particle array for outlet
        source_pa : particle_array
            particle_array of the fluid
        ghost_pa : particle_array
            particle_array of the outlet ghost
        outletinfo : OutletInfo instance
            contains information fo outlet
        kernel : Kernel instance
            Kernel to be used for computations
        dim : int
            dimnesion of the problem
        active_stages : list
            stages of integrator at which update should be active
        callback : function
            callback after the update function
        """
        self.outlet_pa = outlet_pa
        self.source_pa = source_pa
        self.ghost_pa = ghost_pa
        self.dim = dim
        self.kernel = kernel
        self.outletinfo = outletinfo
        self.x = self.y = self.z = 0.0
        self.xn = self.yn = self.zn = 0.0
        self.length = 0.0
        self.callback = callback
        self.active_stages = active_stages
        self.io_eval = None
        self._init = False
        self.props_to_copy = None
        cfg = get_config()
        self.gpu = cfg.use_opencl or cfg.use_cuda

    def initialize(self):
        """Function to initialize the class variables after
        evaluation in SimpleInletOutlet class"""
        outletinfo = self.outletinfo
        self.x = outletinfo.refpoint[0]
        self.y = outletinfo.refpoint[1]
        self.z = outletinfo.refpoint[2]
        self.xn = outletinfo.normal[0]
        self.yn = outletinfo.normal[1]
        self.zn = outletinfo.normal[2]
        self.length = outletinfo.length
        self.props_to_copy = outletinfo.props_to_copy

    def _create_io_eval(self):
        """Evaluator to assign ioid to particles leaving a domain"""
        if self.io_eval is None:
            from pysph.sph.equation import Group
            from pysph.tools.sph_evaluator import SPHEvaluator
            o_name = self.outlet_pa.name
            f_name = self.source_pa.name
            eqns = []
            if self.gpu:
                from pysph.base.gpu_nnps import ZOrderGPUNNPS as NNPS
            else:
                from pysph.base.nnps import LinkedListNNPS as NNPS

            eqns.append(Group(equations=[
                IOEvaluate(
                    o_name, [], x=self.x, y=self.y, z=self.z, xn=self.xn,
                    yn=self.yn, zn=self.zn, maxdist=self.length)],
                real=False, update_nnps=False))

            eqns.append(Group(equations=[
                IOEvaluate(
                    f_name, [], x=self.x, y=self.y, z=self.z, xn=self.xn,
                    yn=self.yn, zn=self.zn,)], real=False, update_nnps=False))

            arrays = [self.outlet_pa] + [self.source_pa]
            io_eval = SPHEvaluator(arrays=arrays, equations=eqns, dim=self.dim,
                                   kernel=self.kernel, nnps_factory=NNPS)
            return io_eval
        else:
            return self.io_eval

    def update(self, time, dt, stage):
        """Update function called after each stage"""
        if not self._init:
            self.initialize()
            self._init = True
        if stage in self.active_stages:
            props_to_copy = self.props_to_copy

            outlet_pa = self.outlet_pa
            source_pa = self.source_pa

            self.io_eval = self._create_io_eval()
            self.io_eval.update()
            self.io_eval.evaluate()

            # adding particles to the destination array.
            if self.gpu:
                source_pa.gpu.pull('ioid')
            io_id = source_pa.ioid
            cond = (io_id == 1)
            all_idx = np.where(cond)[0]
            source_pa.extract_particles(
                all_idx, dest_array=outlet_pa, props=props_to_copy)
            source_pa.remove_particles(all_idx)

            if self.gpu:
                outlet_pa.gpu.pull('ioid')
            io_id = outlet_pa.ioid
            cond = (io_id == 2)
            all_idx = np.where(cond)[0]
            outlet_pa.remove_particles(all_idx)

            if self.callback is not None:
                self.callback(source_pa, outlet_pa)



"""
Functions for advection
#######################
"""

from pysph.sph.equation import Equation
from numpy import cos
from numpy import pi as M_PI


class Advect(Equation):
    def loop(self, d_idx, d_ax, d_ay, d_u, d_v):
        d_ax[d_idx] = d_u[d_idx]
        d_ay[d_idx] = d_v[d_idx]


class MixingVelocityUpdate(Equation):
    def __init__(self, dest, sources, T):
        self.T = T
        super(MixingVelocityUpdate, self).__init__(dest, sources)

    def loop(self, d_idx, d_u, d_v, d_u0, d_v0, t=0.1):
        d_u[d_idx] = cos(M_PI*t/self.T) * d_u0[d_idx]
        d_v[d_idx] = -cos(M_PI*t/self.T) * d_v0[d_idx]



"""2D Kelvin Helmoltz Instability example using TVF. (1 hour)
"""
import numpy

# Particle generator
from pysph.base.utils import get_particle_array
from pysph.base.kernels import WendlandQuintic

# SPH Equations and Group
from pysph.sph.equation import Group

from pysph.sph.wc.viscosity import ClearyArtificialViscosity

from pysph.sph.wc.transport_velocity import SummationDensity, \
    MomentumEquationPressureGradient,\
    SolidWallPressureBC, SolidWallNoSlipBC, SetWallVelocity, \
    StateEquation, MomentumEquationArtificialStress, MomentumEquationViscosity

from pysph.sph.surface_tension import ColorGradientUsingNumberDensity, \
    InterfaceCurvatureFromNumberDensity, ShadlooYildizSurfaceTensionForce,\
    SmoothedColor

from pysph.sph.gas_dynamics.basic import ScaleSmoothingLength

# PySPH solver and application
from pysph.solver.application import Application
from pysph.solver.solver import Solver

# Integrators and Steppers
from pysph.sph.integrator_step import TransportVelocityStep
from pysph.sph.integrator import PECIntegrator

# Domain manager for periodic domains
from pysph.base.nnps import DomainManager

# problem parameters
dim = 2
domain_width = 1.0
domain_height = 1.0

# numerical constants
gy = -9.81
alpha = 0.001
wavelength = 1.0
wavenumber = 2*numpy.pi/wavelength
Ri = 0.05
rho0 = rho1 = 1000.0
rho2 = 1*rho1
U = 0.5
sigma = Ri * (rho1*rho2) * (2*U)**2/(wavenumber*(rho1 + rho2))

# initial perturbation amplitude
psi0 = 0.03*domain_height

# discretization parameters
nghost_layers = 5
dx = dy = 0.01
dxb2 = dyb2 = 0.5 * dx
volume = dx*dx
hdx = 1.5
h0 = hdx * dx
rho0 = 1000.0
c0 = 25.0
p0 = c0*c0*rho0
nu = 0.125 * alpha * h0 * c0

# time steps
tf = 3.0
dt_cfl = 0.25 * h0/(1.1*c0)
dt_viscous = 0.125 * h0**2/nu
dt_force = 1.0

dt = 0.8 * min(dt_cfl, dt_viscous, dt_force)


class KHITVF(Application):
    def create_particles(self):
        ghost_extent = (nghost_layers + 0.5)*dx

        x, y = numpy.mgrid[dxb2:domain_width:dx,
                           -ghost_extent:domain_height+ghost_extent:dy]
        x = x.ravel()
        y = y.ravel()

        m = numpy.ones_like(x) * volume * rho0
        rho = numpy.ones_like(x) * rho0
        h = numpy.ones_like(x) * h0
        cs = numpy.ones_like(x) * c0

        # additional properties required for the fluid.
        additional_props = [
            # volume inverse or number density
            'V',

            # color and gradients
            'color', 'scolor', 'cx', 'cy', 'cz', 'cx2', 'cy2', 'cz2',

            # discretized interface normals and dirac delta
            'nx', 'ny', 'nz', 'ddelta',

            # interface curvature
            'kappa',

            # transport velocities
            'uhat', 'vhat', 'what', 'auhat', 'avhat', 'awhat',

            # imposed accelerations on the solid wall
            'ax', 'ay', 'az', 'wij',

            # velocity of magnitude squared
            'vmag2',

            # variable to indicate reliable normals and normalizing
            # constant
            'N', 'wij_sum',

            ]

        # get the fluid particle array
        fluid = get_particle_array(
            name='fluid', x=x, y=y, h=h, m=m, rho=rho, cs=cs,
            additional_props=additional_props)

        # set the fluid velocity with respect to the sinusoidal
        # perturbation
        fluid.u[:] = -U
        mode = 1
        for i in range(len(fluid.x)):
            ang = 2*numpy.pi*fluid.x[i]/(mode*domain_width)
            temp = domain_height/2 + psi0*domain_height*numpy.sin(ang)
            if fluid.y[i] > temp:
                fluid.u[i] = U
                fluid.color[i] = 1
                fluid.rho[i] = rho1
                fluid.m[i] = volume*rho1
            else:
                fluid.rho[i] = rho2
                fluid.m[i] = rho2/rho1*volume*rho2

        # extract the top and bottom boundary particles
        indices = numpy.where(fluid.y > domain_height)[0]
        wall = fluid.extract_particles(indices)
        fluid.remove_particles(indices)

        indices = numpy.where(fluid.y < 0)[0]
        bottom = fluid.extract_particles(indices)
        fluid.remove_particles(indices)

        # concatenate the two boundaries
        wall.append_parray(bottom)
        wall.set_name('wall')

        # set the number density initially for all particles
        fluid.V[:] = 1./volume
        wall.V[:] = 1./volume

        # set additional output arrays for the fluid
        fluid.add_output_arrays(['V', 'color', 'cx', 'cy', 'nx', 'ny',
                                 'ddelta', 'kappa', 'N', 'p', 'rho'])

        # extrapolated velocities for the wall
        for name in ['uf', 'vf', 'wf']:
            wall.add_property(name)

        # dummy velocities for the wall
        # required for the no-slip BC
        for name in ['ug', 'vg', 'wg']:
            wall.add_property(name)

        print("2D KHI with %d fluid particles and %d wall particles" % (
                fluid.get_number_of_particles(),
                wall.get_number_of_particles()))

        return [fluid, wall]

    def create_domain(self):
        return DomainManager(xmin=0, xmax=domain_width, ymin=0,
                             ymax=domain_height,
                             periodic_in_x=True, periodic_in_y=False)

    def create_solver(self):
        kernel = WendlandQuintic(dim=2)
        integrator = PECIntegrator(fluid=TransportVelocityStep())
        solver = Solver(
            kernel=kernel, dim=dim, integrator=integrator,
            dt=dt, tf=tf, adaptive_timestep=False)
        return solver

    def create_equations(self):
        tvf_equations = [

            # We first compute the mass and number density of the fluid
            # phase. This is used in all force computations henceforth. The
            # number density (1/volume) is explicitly set for the solid phase
            # and this isn't modified for the simulation.
            Group(equations=[
                    SummationDensity(dest='fluid', sources=['fluid', 'wall'])
                    ]),

            # Given the updated number density for the fluid, we can update
            # the fluid pressure. Additionally, we can extrapolate the fluid
            # velocity to the wall for the no-slip boundary
            # condition. Also compute the smoothed color based on the color
            # index for a particle.
            Group(equations=[
                    StateEquation(dest='fluid', sources=None, rho0=rho0,
                                  p0=p0, b=1.0),
                    SetWallVelocity(dest='wall', sources=['fluid']),
                    SmoothedColor(dest='fluid', sources=['fluid']),
                    ]),

            #################################################################
            # Begin Surface tension formulation
            #################################################################
            # Scale the smoothing lengths to determine the interface
            # quantities. The NNPS need not be updated since the smoothing
            # length is decreased.
            Group(equations=[
                    ScaleSmoothingLength(dest='fluid', sources=None,
                                         factor=0.8)
                    ], update_nnps=False),

            # Compute the gradient of the color function with respect to the
            # new smoothing length. At the end of this Group, we will have the
            # interface normals and the discretized dirac delta function for
            # the fluid-fluid interface.
            Group(equations=[
                    ColorGradientUsingNumberDensity(dest='fluid',
                                                    sources=['fluid', 'wall'],
                                                    epsilon=0.01/h0),
                    ],
                  ),

            # Compute the interface curvature using the modified smoothing
            # length and interface normals computed in the previous Group.
            Group(equations=[
                    InterfaceCurvatureFromNumberDensity(
                        dest='fluid',
                        sources=['fluid'],
                        with_morris_correction=True),
                    ], ),

            # Now rescale the smoothing length to the original value for the
            # rest of the computations.
            Group(equations=[
                    ScaleSmoothingLength(dest='fluid', sources=None,
                                         factor=1.25)
                    ], update_nnps=False,
                  ),
            #################################################################
            # End Surface tension formulation
            #################################################################

            # Once the pressure for the fluid phase has been updated via the
            # state-equation, we can extrapolate the pressure to the wall
            # ghost particles. After this group, the density and pressure of
            # the boundary particles has been updated and can be used in the
            # integration equations.
            Group(
                equations=[
                    SolidWallPressureBC(dest='wall', sources=['fluid'], p0=p0,
                                        rho0=rho0, gy=gy),

                    ], ),

            # The main acceleration block
            Group(
                equations=[

                    # Gradient of pressure for the fluid phase using the
                    # number density formulation. No penetration boundary
                    # condition using Adami et al's generalized wall boundary
                    # condition. The extrapolated pressure and density on the
                    # wall particles is used in the gradient of pressure to
                    # simulate a repulsive force.
                    MomentumEquationPressureGradient(
                        dest='fluid', sources=['fluid', 'wall'], pb=p0,
                        gy=gy),

                    # Artificial viscosity for the fluid phase.
                    MomentumEquationViscosity(
                        dest='fluid', sources=['fluid'], nu=nu),

                    # No-slip boundary condition using Adami et al's
                    # generalized wall boundary condition. This equation
                    # basically computes the viscous contribution on the fluid
                    # from the wall particles.
                    SolidWallNoSlipBC(dest='fluid', sources=['wall'], nu=nu),

                    # Surface tension force for the SY11 formulation
                    ShadlooYildizSurfaceTensionForce(dest='fluid',
                                                     sources=None,
                                                     sigma=sigma),

                    # Artificial stress for the fluid phase
                    MomentumEquationArtificialStress(dest='fluid',
                                                     sources=['fluid']),

                    ], )
        ]
        return tvf_equations


if __name__ == '__main__':
    app = KHITVF()
    app.run()



"""GSPH functions"""

from math import sqrt

from compyle.api import declare


def printf(s):
    print(s)


def SIGN(x=0.0, y=0.0):
    if y >= 0:
        return abs(x)
    else:
        return -abs(x)


def riemann_solve(method=1, rhol=0.0, rhor=1.0, pl=0.0, pr=1.0, ul=0.0, ur=1.0,
                  gamma=1.4, niter=20, tol=1e-6, result=[0.0, 0.0]):
    if method == 0:
        return non_diffusive(rhol, rhor, pl, pr, ul, ur, gamma, niter, tol,
                             result)
    elif method == 1:
        return van_leer(rhol, rhor, pl, pr, ul, ur, gamma, niter, tol, result)
    elif method == 2:
        return exact(rhol, rhor, pl, pr, ul, ur, gamma, niter, tol, result)
    elif method == 3:
        return hllc(rhol, rhor, pl, pr, ul, ur, gamma, niter, tol, result)
    elif method == 4:
        return ducowicz(rhol, rhor, pl, pr, ul, ur, gamma, niter, tol, result)
    elif method == 5:
        return hlle(rhol, rhor, pl, pr, ul, ur, gamma, niter, tol, result)
    elif method == 6:
        return roe(rhol, rhor, pl, pr, ul, ur, gamma, niter, tol, result)
    elif method == 7:
        return llxf(rhol, rhor, pl, pr, ul, ur, gamma, niter, tol, result)
    elif method == 8:
        return hllc_ball(rhol, rhor, pl, pr, ul, ur, gamma, niter, tol,
                         result)
    elif method == 9:
        return hll_ball(rhol, rhor, pl, pr, ul, ur, gamma, niter, tol, result)
    elif method == 10:
        return hllsy(rhol, rhor, pl, pr, ul, ur, gamma, niter, tol, result)


def non_diffusive(rhol=0.0, rhor=1.0, pl=0.0, pr=1.0, ul=0.0, ur=1.0,
                  gamma=1.4, niter=20, tol=1e-6, result=[0.0, 0.0]):
    result[0] = 0.5*(pl + pr)
    result[1] = 0.5*(ul + ur)
    return 0


def van_leer(rhol=0.0, rhor=1.0, pl=0.0, pr=1.0,
             ul=0.0, ur=1.0, gamma=1.4, niter=20, tol=1e-6, result=[0.0, 0.0]):
    """Van Leer Riemann solver.


    Parameters
    ----------
    rhol, rhor: double: left and right density.
    pl, pr: double: left and right pressure.
    ul, ur: double: left and right speed.
    gamma: double: Ratio of specific heats.
    niter: int: Max number of iterations to try for iterative schemes.
    tol: double: Error tolerance for convergence.

    result: list: List of length 2. Result will contain computed pstar, ustar

    Returns
    -------

    Returns 0 if the value is computed correctly else 1 if the iterations
    either did not converge or if there is an error.

    """
    if ((rhol < 0) or (rhor < 0) or (pl < 0) or (pr < 0)):
        result[0] = 0.0
        result[1] = 0.0
        return 1

    # local variables
    gamma1, gamma2 = declare('double', 2)
    Vl, Vr, cl, cr, zl, zr, wl, wr = declare('double', 8)
    ustar_l, ustar_r, pstar, pstar_old = declare('double', 4)
    converged, iteration = declare('int', 2)

    gamma2 = 1.0 + gamma
    gamma1 = 0.5 * gamma2/gamma
    smallp = 1e-25

    # initialize variables
    wl = 0.0
    wr = 0.0
    zl = 0.0
    zr = 0.0

    # specific volumes
    Vl = 1./rhol
    Vr = 1./rhor

    # Lagrangean sound speeds
    cl = sqrt(gamma * pl * rhol)
    cr = sqrt(gamma * pr * rhor)

    # Initial guess for pstar
    pstar = pr - pl - cr * (ur - ul)
    pstar = pl + pstar * cl/(cl + cr)

    pstar = max(pstar, smallp)

    # Now Iterate using NR to obtain the star values
    iteration = 0
    while iteration < niter:
        pstar_old = pstar

        wl = 1.0 + gamma1 * (pstar - pl)/pl
        wl = cl * sqrt(wl)

        wr = 1.0 + gamma1 * (pstar - pr)/pr
        wr = cr * sqrt(wr)

        zl = 4.0 * Vl * wl * wl
        zl = -zl * wl/(zl - gamma2*(pstar - pl))

        zr = 4.0 * Vr * wr * wr
        zr = zr * wr/(zr - gamma2*(pstar - pr))

        ustar_l = ul - (pstar - pl)/wl
        ustar_r = ur + (pstar - pr)/wr

        pstar = pstar + (ustar_r - ustar_l)*(zl*zr)/(zr - zl)
        pstar = max(smallp, pstar)

        converged = (abs(pstar - pstar_old)/pstar < tol)
        if converged:
            break

        iteration += 1

    # calculate the averaged ustar
    ustar_l = ul - (pstar - pl)/wl
    ustar_r = ur + (pstar - pr)/wr
    ustar = 0.5 * (ustar_l + ustar_r)

    result[0] = pstar
    result[1] = ustar
    if converged:
        return 0
    else:
        return 1


def prefun_exact(p=0.0, dk=0.0, pk=0.0, ck=0.0, g1=0.0, g2=0.0,
                 g4=0.0, g5=0.0, g6=0.0, result=[0.0, 0.0]):
    """The pressure function.  Updates result with f, fd.
    """

    pratio, f, fd, ak, bk, qrt = declare('double', 6)

    if (p <= pk):
        pratio = p/pk
        f = g4*ck*(pratio**g1 - 1.0)
        fd = (1.0/(dk*ck))*pratio**(-g2)
    else:
        ak = g5/dk
        bk = g6*pk
        qrt = sqrt(ak/(bk+p))
        f = (p-pk)*qrt
        fd = (1.0 - 0.5*(p-pk)/(bk + p))*qrt

    result[0] = f
    result[1] = fd


def exact(rhol=0.0, rhor=1.0, pl=0.0, pr=1.0,
          ul=0.0, ur=1.0, gamma=1.4, niter=20, tol=1e-6, result=[0.0, 0.0]):
    """Exact Riemann solver.

    Solve the Riemann problem for the Euler equations to determine the
    intermediate states.

    Parameters
    ----------
    rhol, rhor: double: left and right density.
    pl, pr: double: left and right pressure.
    ul, ur: double: left and right speed.
    gamma: double: Ratio of specific heats.
    niter: int: Max number of iterations to try for iterative schemes.
    tol: double: Error tolerance for convergence.

    result: list: List of length 2. Result will contain computed pstar, ustar

    Returns
    -------

    Returns 0 if the value is computed correctly else 1 if the iterations
    either did not converge or if there is an error.

    """
    # save derived variables
    tmp1 = 1.0/(2*gamma)
    tmp2 = 1.0/(gamma - 1.0)
    tmp3 = 1.0/(gamma + 1.0)

    gamma1 = (gamma - 1.0) * tmp1
    gamma2 = (gamma + 1.0) * tmp1
    gamma3 = 2*gamma * tmp2
    gamma4 = 2 * tmp2
    gamma5 = 2 * tmp3
    gamma6 = tmp3/tmp2
    gamma7 = 0.5 * (gamma - 1.0)

    # sound speed to the left and right based on the Ideal EOS
    cl, cr = declare('double', 2)
    cl = sqrt(gamma*pl/rhol)
    cr = sqrt(gamma*pr/rhor)

    # Iteration constants
    i = declare('int')
    i = 0

    # local variables
    qser, cup, ppv, pmin, pmax, qmax = declare('double', 6)
    pq, um, ptl, ptr, pm, gel, ger = declare('double', 7)
    pstar, pold, udifff = declare('double', 3)
    p, change = declare('double', 2)

    # initialize variables
    fl, fr = declare('matrix(2)', 2)
    p = 0.0

    # check the initial data
    if (gamma4*(cl+cr) <= (ur-ul)):
        return 1

    # compute the guess pressure 'pm'
    qser = 2.0
    cup = 0.25 * (rhol + rhor)*(cl+cr)
    ppv = 0.5 * (pl + pr) + 0.5*(ul - ur)*cup
    pmin = min(pl, pr)
    pmax = max(pl, pr)
    qmax = pmax/pmin

    ppv = max(0.0, ppv)

    if ((qmax <= qser) and (pmin <= ppv) and (ppv <= pmax)):
        pm = ppv
    elif (ppv < pmin):
        pq = (pl/pr)**gamma1
        um = (pq*ul/cl + ur/cr + gamma4*(pq - 1.0))/(pq/cl + 1.0/cr)
        ptl = 1.0 + gamma7 * (ul - um)/cl
        ptr = 1.0 + gamma7 * (um - ur)/cr
        pm = 0.5*(pl*ptl**gamma3 + pr*ptr**gamma3)
    else:
        gel = sqrt((gamma5/rhol)/(gamma6*pl + ppv))
        ger = sqrt((gamma5/rhor)/(gamma6*pr + ppv))
        pm = (gel*pl + ger*pr - (ur-ul))/(gel + ger)

    # the guessed value is pm
    pstart = pm

    pold = pstart
    udifff = ur-ul

    for i in range(niter):
        prefun_exact(pold, rhol, pl, cl, gamma1, gamma2,
                     gamma4, gamma5, gamma6, fl)
        prefun_exact(pold, rhor, pr, cr, gamma1, gamma2,
                     gamma4, gamma5, gamma6, fr)

        p = pold - (fl[0] + fr[0] + udifff)/(fl[1] + fr[1])
        change = 2.0 * abs((p - pold)/(p + pold))
        if change <= tol:
            break
        pold = p

    if i == niter - 1:
        printf(b"%s", b"Divergence in Newton-Raphson Iteration")
        return 1

    # compute the velocity in the star region 'um'
    um = 0.5 * (ul + ur + fr[0] - fl[0])
    result[0] = p
    result[1] = um
    return 0


def sample(pm=0.0, um=0.0, s=0.0, rhol=1.0, rhor=0.0, pl=1.0, pr=0.0,
           ul=1.0, ur=0.0, gamma=1.4, result=[0.0, 0.0, 0.0]):
    """Sample the solution to the Riemann problem.

    Parameters
    ----------

    pm, um : float
        Pressure and velocity in the star region as returned
        by `exact`

    s : float
        Sampling point (x/t)

    rhol, rhor : float
        Densities on either side of the discontinuity

    pl, pr : float
        Pressures on either side of the discontinuity

    ul, ur : float
        Velocities on either side of the discontinuity

    gamma : float
        Ratio of specific heats

    result : list(double)
        (rho, u, p) : Sampled density, velocity and pressure

    """
    tmp1, tmp2, tmp3 = declare('double', 3)
    g1, g2, g3, g4, g5, g6 = declare('double', 6)
    cl, cr = declare('double', 2)

    # save derived variables
    tmp1 = 1.0/(2*gamma)
    tmp2 = 1.0/(gamma - 1.0)
    tmp3 = 1.0/(gamma + 1.0)

    g1 = (gamma - 1.0) * tmp1
    g2 = (gamma + 1.0) * tmp1
    g3 = 2*gamma * tmp2
    g4 = 2 * tmp2
    g5 = 2 * tmp3
    g6 = tmp3/tmp2
    g7 = 0.5 * (gamma - 1.0)

    # sound speeds at the left and right data states
    cl = sqrt(gamma*pl/rhol)
    cr = sqrt(gamma*pr/rhor)

    if s <= um:
        # sampling point lies to the left of the contact discontinuity
        if (pm <= pl):  # left rarefaction
            # speed of the head of the rarefaction
            shl = ul - cl

            if (s <= shl):
                # sampled point is left state
                rho = rhol
                u = ul
                p = pl
            else:
                cml = cl*(pm/pl)**g1    # eqn (4.54)
                stl = um - cml          # eqn (4.55)

                if (s > stl):
                    # sampled point is Star left state. eqn (4.53)
                    rho = rhol*(pm/pl)**(1.0/gamma)
                    u = um
                    p = pm
                else:
                    # sampled point is inside left fan
                    u = g5 * (cl + g7*ul + s)
                    c = g5 * (cl + g7*(ul - s))

                    rho = rhol*(c/cl)**g4
                    p = pl * (c/cl)**g3

        else:  # pm <= pl
            # left shock
            pml = pm/pl
            sl = ul - cl*sqrt(g2*pml + g1)

            if (s <= sl):
                # sampled point is left data state
                rho = rhol
                u = ul
                p = pl
            else:
                # sampled point is Star left state
                rho = rhol*(pml + g6)/(pml*g6 + 1.0)
                u = um
                p = pm

    else:  # s < um
        # sampling point lies to the right of the contact discontinuity
        if (pm > pr):
            # right shock
            pmr = pm/pr
            sr = ur + cr * sqrt(g2*pmr + g1)

            if (s >= sr):
                # sampled point is right data state
                rho = rhor
                u = ur
                p = pr
            else:
                # sampled point is star right state
                rho = rhor*(pmr + g6)/(pmr*g6 + 1.0)
                u = um
                p = pm
        else:
            # right rarefaction
            shr = ur + cr

            if (s >= shr):
                # sampled point is right state
                rho = rhor
                u = ur
                p = pr
            else:
                cmr = cr*(pm/pr)**g1
                STR = um + cmr

                if (s <= STR):
                    # sampled point is star right
                    rho = rhor*(pm/pr)**(1.0/gamma)
                    u = um
                    p = pm
                else:
                    # sampled point is inside left fan
                    u = g5*(-cr + g7*ur + s)
                    c = g5*(cr - g7*(ur - s))
                    rho = rhor * (c/cr)**g4
                    p = pr*(c/cr)**g3

    result[0] = rho
    result[1] = u
    result[2] = p


def ducowicz(rhol=0.0, rhor=1.0, pl=0.0, pr=1.0,
             ul=0.0, ur=1.0, gamma=1.4, niter=20, tol=1e-6, result=[0.0, 0.0]):
    """Ducowicz Approximate Riemann solver for the Euler equations to
    determine the intermediate states.

    Parameters
    ----------
    rhol, rhor: double: left and right density.
    pl, pr: double: left and right pressure.
    ul, ur: double: left and right speed.
    gamma: double: Ratio of specific heats.
    niter: int: Max number of iterations to try for iterative schemes.
    tol: double: Error tolerance for convergence.

    result: list: List of length 2. Result will contain computed pstar, ustar

    Returns
    -------

    Returns 0 if the value is computed correctly else 1 if the iterations
    either did not converge or if there is an error.

    """
    al, ar = declare('double', 2)
    # strong shock parameters
    al = 0.5 * (gamma + 1.0)
    ar = 0.5 * (gamma + 1.0)

    csl, csr, umin, umax, plmin, prmin, bl, br = declare('double', 8)
    a, b, c, d, pstar, ustar, dd = declare('double', 7)
    # Lagrangian sound speeds
    csl = sqrt(gamma * pl * rhol)
    csr = sqrt(gamma * pr * rhor)

    umin = ur - 0.5 * csr/ar
    umax = ul + 0.5 * csl/al

    plmin = pl - 0.25*rhol*csl*csl/al
    prmin = pr - 0.25*rhor*csr*csr/ar

    bl = rhol*al
    br = rhor*ar

    a = (br-bl) * (prmin-plmin)
    b = br*umin*umin - bl*umax*umax
    c = br*umin - bl*umax
    d = br*bl * (umin-umax)*(umin-umax)

    # Case A : ustar - umin > 0
    dd = sqrt(max(0.0, d-a))
    ustar = (b + prmin-plmin)/(c - SIGN(dd, umax-umin))

    if (((ustar - umin) >= 0.0) and ((ustar - umax) <= 0)):
        pstar = 0.5 * (plmin + prmin + br*abs(ustar-umin)*(ustar-umin) -
                       bl*abs(ustar-umax)*(ustar-umax))
        pstar = max(pstar, 0.0)
        result[0] = pstar
        result[1] = ustar
        return 0

    # Case B: ustar - umin < 0, ustar - umax > 0
    dd = sqrt(max(0.0, d+a))
    ustar = (b - prmin + plmin)/(c - SIGN(dd, umax-umin))
    if (((ustar-umin) <= 0.0) and ((ustar-umax) >= 0.0)):
        pstar = 0.5 * (plmin+prmin + br*abs(ustar-umin)*(ustar-umin) -
                       bl*abs(ustar-umax)*(ustar-umax))
        pstar = max(pstar, 0.0)
        result[0] = pstar
        result[1] = ustar
        return 0

    a = (bl+br)*(plmin-prmin)
    b = bl*umax + br*umin
    c = 1./(bl + br)

    # Case C : ustar-umin >0, ustar-umax > 0
    dd = sqrt(max(0.0, a-d))
    ustar = (b+dd)*c
    if (((ustar-umin) >= 0) and ((ustar-umax) >= 0.0)):
        pstar = 0.5 * (plmin+prmin + br*abs(ustar-umin)*(ustar-umin)
                       - bl*abs(ustar-umax)*(ustar-umax))
        pstar = max(pstar, 0.0)
        result[0] = pstar
        result[1] = ustar
        return 0

    # Case D: ustar - umin < 0, ustar - umax < 0
    dd = sqrt(max(0.0, -a - d))
    ustar = (b-dd)*c
    pstar = 0.5 * (plmin+prmin + br*abs(ustar-umin)*(ustar-umin)
                   - bl*abs(ustar-umax)*(ustar-umax))
    pstar = max(pstar, 0.0)
    result[0] = pstar
    result[1] = ustar
    return 0


def roe(rhol=0.0, rhor=1.0, pl=0.0, pr=1.0, ul=0.0, ur=1.0,
        gamma=1.4, niter=20, tol=1e-6, result=[0.0, 0.0]):
    """Roe's approximate Riemann solver for the Euler equations to
    determine the intermediate states.

    Parameters
    ----------
    rhol, rhor: double: left and right density.
    pl, pr: double: left and right pressure.
    ul, ur: double: left and right speed.
    gamma: double: Ratio of specific heats.
    niter: int: Max number of iterations to try for iterative schemes.
    tol: double: Error tolerance for convergence.

    result: list: List of length 2. Result will contain computed pstar, ustar

    Returns
    -------

    Returns 0 if the value is computed correctly else 1 if the iterations
    either did not converge or if there is an error.

    """

    rrhol, rrhor, denominator, plr, vlr, ulr = declare('double', 6)
    cslr, cslr1 = declare('double', 2)
    # Roe-averaged pressure and specific volume
    rrhol = sqrt(rhol)
    rrhor = sqrt(rhor)

    denominator = 1./(rrhor + rrhol)

    plr = (rrhol*pl + rrhor*pr) * denominator
    vlr = (rrhol/rhol + rrhor/rhor) * denominator
    ulr = (rrhol*ul + rrhor*ur) * denominator

    # average sound speed at the interface
    cslr = sqrt(gamma * plr/vlr)
    cslr1 = 1./cslr

    # the intermediate states
    result[0] = plr - 0.5 * (ur - ul) * cslr
    result[1] = ulr - 0.5 * (pr - pl) * cslr1

    return 0


def llxf(rhol=0.0, rhor=1.0, pl=0.0, pr=1.0, ul=0.0, ur=1.0,
         gamma=1.4, niter=20, tol=1e-6, result=[0.0, 0.0]):
    """Lax Friedrichs approximate Riemann solver for the Euler equations to
    determine the intermediate states.

    Parameters
    ----------
    rhol, rhor: double: left and right density.
    pl, pr: double: left and right pressure.
    ul, ur: double: left and right speed.
    gamma: double: Ratio of specific heats.
    niter: int: Max number of iterations to try for iterative schemes.
    tol: double: Error tolerance for convergence.

    result: list: List of length 2. Result will contain computed pstar, ustar

    Returns
    -------

    Returns 0 if the value is computed correctly else 1 if the iterations
    either did not converge or if there is an error.

    """

    gamma1, csl, csr, cslr, el, El, er, Er, pstar = declare('double', 9)
    gamma1 = 1./(gamma - 1.0)

    # Lagrangian sound speeds
    csl = sqrt(gamma * pl * rhol)
    csr = sqrt(gamma * pr * rhor)
    cslr = max(csr, csl)

    # Total energy on either side
    el = pl*gamma1/rhol
    El = el + 0.5 * ul*ul

    er = pr*gamma1/rhor
    Er = er + 0.5 * ur*ur

    # the intermediate states
    # cdef double ustar = 0.5 * ( ul + ur - 1./cslr * (pr - pl) )
    pstar = 0.5 * (pl + pr - cslr * (ur - ul))
    result[0] = pstar
    result[1] = 1./pstar * (0.5 * ((pl*ul + pr*ur) - cslr*(Er - El)))
    return 0


def hllc(rhol=0.0, rhor=1.0, pl=0.0, pr=1.0, ul=0.0, ur=1.0,
         gamma=1.4, niter=20, tol=1e-6, result=[0.0, 0.0]):
    """Harten-Lax-van Leer-Contact approximate Riemann solver for the Euler
    equations to determine the intermediate states.

    Parameters
    ----------
    rhol, rhor: double: left and right density.
    pl, pr: double: left and right pressure.
    ul, ur: double: left and right speed.
    gamma: double: Ratio of specific heats.
    niter: int: Max number of iterations to try for iterative schemes.
    tol: double: Error tolerance for convergence.

    result: list: List of length 2. Result will contain computed pstar, ustar

    Returns
    -------

    Returns 0 if the value is computed correctly else 1 if the iterations
    either did not converge or if there is an error.

    """

    gamma1, pstar, ustar, mstar, estar = declare('double', 5)
    gamma1 = 1./(gamma - 1.0)

    # Roe-averaged interface velocity
    rrhol = sqrt(rhol)
    rrhor = sqrt(rhor)
    ulr = (rrhol*ul + rrhor*ur)/(rrhol + rrhor)

    # relative velocities at the interface
    vl = ul - ulr
    vr = ur - ulr

    # Sound speeds at the interface
    csl = sqrt(gamma * pl/rhol)
    csr = sqrt(gamma * pr/rhor)
    cslr = (rrhol*csl + rrhor*csr)/(rrhol + rrhor)

    # wave speed estimates at the interface
    sl = min(vl - csl, 0 - cslr)
    sr = max(vr + csr, 0 + cslr)

    sm = rhor*vr*(sr-vr) - rhol*vl*(sl-vl) + pl - pr
    sm /= (rhor*(sr-vr) - rhol*(sl-vl))

    # phat
    phat = rhol*(vl - sl)*(vl - sm) + pl

    # Total energy on either side
    el = pl*gamma1/rhol
    El = rhol*(el + 0.5 * ul*ul)

    er = pr*gamma1/rhor
    Er = rhor*(er + 0.5 * ur*ur)

    # Momentum on either side
    Ml = rhol*ul
    Mr = rhor*ur

    # compute the values based on the wave speeds
    if (sl > 0):
        pstar = pl
        ustar = ul

    elif (sl <= 0.0) and (0.0 < sm):
        mstar = 1./(sl - sm) * ((sl - vl) * Ml + (phat - pl))
        estar = 1./(sl - sm) * ((sl - vl) * El - pl*vl + phat*sm)

        pstar = sm*mstar + phat

        ustar = sm*estar + (sm + ulr)*phat
        ustar /= pstar

    elif (sm <= 0) and (0 < sr):
        mstar = 1./(sr - sm) * ((sr - vr) * Mr + (phat - pr))
        estar = 1./(sr - sm) * ((sr - vr) * Er - pr*vr + phat*sm)

        pstar = sm*mstar + phat

        ustar = sm*estar + (sm + ulr)*phat
        ustar /= pstar

    elif (sr < 0):
        pstar = pr
        ustar = ur

    else:
        printf(b"%s", b"Incorrect wave speeds")
        return 1

    result[0] = pstar
    result[1] = ustar
    return 0


def hllc_ball(rhol=0.0, rhor=1.0, pl=0.0, pr=1.0, ul=0.0, ur=1.0,
              gamma=1.4, niter=20, tol=1e-6, result=[0.0, 0.0]):
    """Harten-Lax-van Leer-Contact approximate Riemann solver for the Euler
    equations to determine the intermediate states.

    Parameters
    ----------
    rhol, rhor: double: left and right density.
    pl, pr: double: left and right pressure.
    ul, ur: double: left and right speed.
    gamma: double: Ratio of specific heats.
    niter: int: Max number of iterations to try for iterative schemes.
    tol: double: Error tolerance for convergence.

    result: list: List of length 2. Result will contain computed pstar, ustar

    Returns
    -------

    Returns 0 if the value is computed correctly else 1 if the iterations
    either did not converge or if there is an error.

    """
    gamma1, csl, csr, cslr, rholr, pstar, ustar = declare('double', 7)
    Sl, Sr, ql, qr, pstar_l, pstar_r = declare('double', 6)

    gamma1 = 0.5 * (gamma + 1.0)/gamma

    # sound speeds in the undisturbed fluid
    csl = sqrt(gamma * pl/rhol)
    csr = sqrt(gamma * pr/rhor)
    cslr = 0.5 * (csl + csr)

    # averaged density
    rholr = 0.5 * (rhol + rhor)

    # provisional intermediate pressure
    pstar = 0.5 * (pl + pr - rholr*cslr*(ur - ul))

    # Wave speed estimates (ustar is taken as the intermediate velocity)
    ustar = 0.5 * (ul + ur - 1./(rholr*cslr)*(pr - pl))

    Hl = pstar/pl
    Hr = pstar/pr

    ql = 1.0
    if (Hl > 1):
        ql = sqrt(1 + gamma1*(Hl - 1.0))

    qr = 1.0
    if (Hr > 1):
        qr = sqrt(1 + gamma1*(Hr - 1.0))

    Sl = ul - csl*ql
    Sr = ur + csr*qr

    # compute the intermediate pressure
    pstar_l = pl + rhol*(ul - Sl)*(ul - ustar)
    pstar_r = pr + rhor*(ur - Sr)*(ur - ustar)

    pstar = 0.5 * (pstar_l + pstar_r)

    # return intermediate states
    result[0] = pstar
    result[1] = ustar
    return 0


def hlle(rhol=0.0, rhor=1.0, pl=0.0, pr=1.0, ul=0.0, ur=1.0,
         gamma=1.4, niter=20, tol=1e-6, result=[0.0, 0.0]):
    """Harten-Lax-van Leer-Einfeldt approximate Riemann solver for the Euler
    equations to determine the intermediate states.

    Parameters
    ----------
    rhol, rhor: double: left and right density.
    pl, pr: double: left and right pressure.
    ul, ur: double: left and right speed.
    gamma: double: Ratio of specific heats.
    niter: int: Max number of iterations to try for iterative schemes.
    tol: double: Error tolerance for convergence.

    result: list: List of length 2. Result will contain computed pstar, ustar

    Returns
    -------

    Returns 0 if the value is computed correctly else 1 if the iterations
    either did not converge or if there is an error.

    """
    gamma1 = 1./(gamma - 1.0)

    # Roe-averaged interface velocity
    rrhol = sqrt(rhol)
    rrhor = sqrt(rhor)
    # ulr = (rrhol*ul + rrhor*ur)/(rrhol + rrhor)

    # lagrangian sound speeds
    csl = sqrt(gamma * pl * rhol)
    csr = sqrt(gamma * pr * rhor)
    cslr = (rrhol*csl + rrhor*csr)/(rrhol + rrhor)

    # wave speed estimates
    sl = min(ul - csl, -cslr)
    sr = max(ur + csr, +cslr)

    smax = max(sl, sr)
    smin = min(sl, sr)
    # cdef double smax = max( (ur-ulr) + csr,  cslr )
    # cdef double smin = max( (ul-ulr) - csl, -cslr )

    # Total energy on either side
    el = pl*gamma1/rhol
    El = el + 0.5 * ul*ul

    er = pr*gamma1/rhor
    Er = er + 0.5 * ur*ur

    # Momentum on either side
    # Ml = rhol*ul
    # Mr = rhor*ur

    # the intermediate states
    pstar = ((smax * pl - smin * pr)/(smax - smin) +
             smax*smin/(smax - smin)*(ur - ul))
    ustar = ((smax * pl*ul - smin * pr*ur)/(smax - smin) +
             smax*smin/(smax - smin)*(Er - El))
    result[0] = pstar
    result[1] = ustar/pstar

    return 0


def hll_ball(rhol=0.0, rhor=1.0, pl=0.0, pr=1.0, ul=0.0, ur=1.0,
             gamma=1.4, niter=20, tol=1e-6, result=[0.0, 0.0]):
    """Harten-Lax-van Leer-Ball approximate Riemann solver for the Euler equations
    to determine the intermediate states.

    Parameters
    ----------
    rhol, rhor: double: left and right density.
    pl, pr: double: left and right pressure.
    ul, ur: double: left and right speed.
    gamma: double: Ratio of specific heats.
    niter: int: Max number of iterations to try for iterative schemes.
    tol: double: Error tolerance for convergence.

    result: list: List of length 2. Result will contain computed pstar, ustar

    Returns
    -------

    Returns 0 if the value is computed correctly else 1 if the iterations
    either did not converge or if there is an error.

    """
    # Roe-averaged pressure and specific volume
    rrhol = sqrt(rhol)
    rrhor = sqrt(rhor)

    denominator = 1./(rrhor + rrhol)

    # sound speeds
    csl = sqrt(gamma * pl/rhol)
    csr = sqrt(gamma * pr/rhor)
    # cslr2 = denominator * (rrhol*csl*csl + rrhor*csr*csr )
    eta = 0.5 * (gamma - 1.0) * (rrhor*rrhol) * denominator * denominator
    betal = abs(ul)
    betar = abs(ur)

    # averaged velocity and cs2
    ulr = (rrhol*ul + rrhor*ur)/(rrhol*rrhor)
    cslr2 = (rrhol*csl*csl + rrhor*csr*csr)/(rrhol*rrhor)
    cslr = sqrt(cslr2 + eta * (betar - betal) * (betar - betal))

    # wave speed estimates
    Sl = min(ulr - cslr, ul - csl)
    Sr = max(ulr + cslr, ur + csr)

    # intermediate states
    # rhostar = 1./(Sr - Sl) * (rhor*(Sr-ur) - rhol*(Sl-ul))
    ustar = ((Sr*Sl*(rhor-rhol) + rhol*ul*Sr - rhor*ur*Sl) /
             (rhol*(ul-Sl) + rhor*(Sr-ur)))
    # pstar = 0.5 * (pl + pr - rhostar*ustar*( (Sr-ustar) + (Sl-ustar) ) + \
    #                                rhor*ur*(ur - Sr) + rhol*ul*(ul-Sl) )
    pstar = (pr*(ustar-Sl) - pl*(ustar-Sr) +
             rhor*ur*(ustar-Sl)*(ur-Sr) -
             rhol*ul*(ustar-Sr)*(ul-Sl))
    pstar = pstar/(Sr - Sl)
    result[0] = pstar
    result[1] = ustar

    return 0


def hllsy(rhol=0.0, rhor=1.0, pl=0.0, pr=1.0, ul=0.0, ur=1.0,
          gamma=1.4, niter=20, tol=1e-6, result=[0.0, 0.0]):
    """HLL Riemann solver defined by Sirotkin and Yoh in 'A Smoothed Particle
    Hydrodynamics method with approximate Riemann solvers for simulation of
    strong explosions' (2013), Computers and Fluids.

    Parameters
    ----------
    rhol, rhor: double: left and right density.
    pl, pr: double: left and right pressure.
    ul, ur: double: left and right speed.
    gamma: double: Ratio of specific heats.
    niter: int: Max number of iterations to try for iterative schemes.
    tol: double: Error tolerance for convergence.

    result: list: List of length 2. Result will contain computed pstar, ustar

    Returns
    -------

    Returns 0 if the value is computed correctly else 1 if the iterations
    either did not converge or if there is an error.

    """

    gamma1 = 1./(gamma - 1.0)

    # Roe-averaging factors
    rrhol = sqrt(rhol)
    rrhor = sqrt(rhor)
    denominator = 1./(rrhor + rrhol)

    # Lagrangian sound speed Eq. (35) in SY13
    csl = sqrt(gamma * pl*rhol)
    csr = sqrt(gamma * pr*rhor)
    cslr = denominator * (rrhol*csl + rrhor*csr)

    # weighting factors Eqs. (33 - 35) in SY13
    bl = max(csl, cslr)
    br = max(csr, cslr)
    wl = br/(bl + br)
    wr = bl/(bl + br)
    wlr = bl*br/(bl + br)

    # Total energy on either side
    el = pl*gamma1/rhol
    El = el + 0.5 * ul*ul

    er = pr*gamma1/rhor
    Er = er + 0.5 * ur*ur

    # intermediate states Eq.(32) in SY13
    pstar = wl*pl + wr*pr - wlr*(ur - ul)
    ustar = wl*(pl*ul) + wr*(pr*ur) - wlr*(Er - El)
    result[0] = pstar
    result[1] = ustar/pstar
    return 0


HELPERS = [
    SIGN, riemann_solve, non_diffusive,
    ducowicz, exact, hll_ball, hllc,
    hllc_ball, hlle, hllsy, llxf, roe,
    van_leer, prefun_exact
]



"""A convenience class that combines an AccelerationEval and an SPHCompiler to
allow a user to specify particle arrays, equations, an optional domain and
kernel to produce an SPH evaluation.

This is handy for post-processing.

"""

from pysph.base.kernels import Gaussian
from pysph.base.nnps import LinkedListNNPS as NNPS
from pysph.sph.acceleration_eval import AccelerationEval
from pysph.sph.sph_compiler import SPHCompiler


class SPHEvaluator(object):
    def __init__(self, arrays, equations, dim, kernel=None,
                 domain_manager=None, backend=None, nnps_factory=NNPS):
        """Constructor.

        Parameters
        ----------
        arrays: list(ParticleArray)
        equations: list
        dim: int
        kernel: kernel instance.
        domain_manager: DomainManager
        backend: str: indicates the backend to use.
            one of ('opencl', 'cython', '', None)
        nnps_factory: A factory that creates an NNPSBase instance.
        """
        self.arrays = arrays
        self.equations = equations
        self.domain_manager = domain_manager
        self.dim = dim
        if kernel is None:
            self.kernel = Gaussian(dim=dim)
        else:
            self.kernel = kernel

        self.nnps_factory = nnps_factory
        self.backend = backend

        self.func_eval = AccelerationEval(arrays, equations, self.kernel,
                                          backend=backend)
        compiler = SPHCompiler(self.func_eval, None)
        compiler.compile()
        self._create_nnps(arrays)

    def evaluate(self, t=0.0, dt=0.1):
        """Evalute the SPH equations, dummy t and dt values can
        be passed.
        """
        self.func_eval.compute(t, dt)

    def update(self, update_domain=True):
        """Update the NNPS when particles have moved.

        If the update_domain is False, the domain is not updated.

        Use this when the arrays are the same but the particles have themselves
        changed. If the particle arrays themselves change use the
        `update_particle_arrays` method instead.
        """
        if update_domain:
            self.nnps.update_domain()
        self.nnps.update()

    def update_particle_arrays(self, arrays):
        """Call this for a new set of particle arrays which have the
        same properties as before.

        For example, if you are reading the particle array data from files,
        each time you load a new file a new particle array is read with the
        same properties.  Call this function to reset the arrays.
        """
        self._create_nnps(arrays)
        self.func_eval.update_particle_arrays(arrays)

    # Private protocol ###################################################
    def _create_nnps(self, arrays):
        self.nnps = self.nnps_factory(
            dim=self.kernel.dim, particles=arrays,
            radius_scale=self.kernel.radius_scale,
            domain=self.domain_manager, cache=True
        )
        self.func_eval.set_nnps(self.nnps)



"""Different configurations for the 2D Riemann problems.

The different configurations and expected solutions to these problems
are defined in 'Solution of Two Dimensional Riemann Problems for Gas
Dynamics without Riemann Problem Solvers' by Alexander Kurganov and
Eitan Tadmor

General notations for the different configurations are (S) for shock
waves, (R) for Rarefactions and (J) for contact/slip lines

Code from https://bitbucket.org/kunalp/sph2d/
path: src/examples/r2d_config.py
"""


class R2DConfig(object):
    def __init__(self, config=3):
        self.config = config

        self.xmin = -0.25
        self.xmax = 1.15
        self.ymin = -0.25
        self.ymax = 1.15
        self.zmin = 0
        self.zmax = 0

        self.endtime = 0.25

        if config == 12:
            self.setup_config12()

        elif config == 2:
            self.setup_config2()

        elif config == 3:
            self.setup_config3()

        elif config == 4:
            self.setup_config4()

        elif config == 5:
            self.setup_config5()

        elif config == 6:
            self.setup_config6()

        elif config == 8:
            self.setup_config8()

        self.xmid = 0.5 * (self.xmin + self.xmax)
        self.ymid = 0.5 * (self.ymin + self.ymax)

        self.rho_max = max(self.rho1, self.rho2, self.rho3, self.rho4)
        self.rho_min = min(self.rho1, self.rho2, self.rho3, self.rho4)

    def setup_config3(self):
        """Four Shocks"""
        self.endtime = 0.3

        self.p1 = 1.5
        self.rho1 = 1.5
        self.u1 = 0.0
        self.v1 = 0.0

        self.p2 = 0.3
        self.rho2 = 0.5323
        self.u2 = 1.206
        self.v2 = 0.0

        self.p3 = 0.029
        self.rho3 = 0.138
        self.u3 = 1.206
        self.v3 = 1.206

        self.p4 = 0.3
        self.rho4 = 0.5323
        self.u4 = 0.0
        self.v4 = 1.206

    def setup_config2(self):
        """Four Rarefactions"""
        self.endtime = 0.2

        self.p1 = 1.0
        self.rho1 = 1.0
        self.u1 = 0.0
        self.v1 = 0.0

        self.p2 = 0.4
        self.rho2 = 0.5197
        self.u2 = -0.7259
        self.v2 = 0.0

        self.p3 = 1.0
        self.rho3 = 1.0
        self.u3 = -0.7259
        self.v3 = -0.7259

        self.p4 = 0.4
        self.rho4 = 0.5197
        self.u4 = 0.0
        self.v4 = -0.7259

    def setup_config4(self):
        self.endtime = 0.25

        self.p1 = 1.1
        self.rho1 = 1.1
        self.u1 = 0.0
        self.v1 = 0.0

        self.p2 = 0.35
        self.rho2 = 0.5065
        self.u2 = 0.8939
        self.v2 = 0.0

        self.p3 = 1.1
        self.rho3 = 1.1
        self.u3 = 0.8939
        self.v3 = 0.8939

        self.p4 = 0.35
        self.rho4 = 0.5065
        self.u4 = 0.0
        self.v4 = 0.8939

    def setup_config5(self):
        self.endtime = 0.23

        self.p1 = 1
        self.rho1 = 1
        self.u1 = -0.75
        self.v1 = -0.5

        self.p2 = 1.0
        self.rho2 = 2.0
        self.u2 = -0.75
        self.v2 = 0.5

        self.p3 = 1
        self.rho3 = 1
        self.u3 = 0.75
        self.v3 = 0.5

        self.p4 = 1.0
        self.rho4 = 3.0
        self.u4 = 0.75
        self.v4 = -0.5

    def setup_config6(self):
        self.endtime = 0.3

        self.p1 = 1
        self.rho1 = 1
        self.u1 = 0.75
        self.v1 = -0.5

        self.p2 = 1.0
        self.rho2 = 2.0
        self.u2 = 0.75
        self.v2 = 0.5

        self.p3 = 1
        self.rho3 = 1
        self.u3 = -0.75
        self.v3 = 0.5

        self.p4 = 1.0
        self.rho4 = 3.0
        self.u4 = -0.75
        self.v4 = -0.5

    def setup_config8(self):
        self.endtime = 0.25

        self.p1 = 0.4
        self.rho1 = 0.5197
        self.u1 = 0.1
        self.v1 = 0.1

        self.p2 = 1
        self.rho2 = 1.0
        self.u2 = -0.6259
        self.v2 = 0.1

        self.p3 = 1
        self.rho3 = 0.8
        self.u3 = 0.1
        self.v3 = 0.1

        self.p4 = 1.0
        self.rho4 = 1.0
        self.u4 = 0.1
        self.v4 = -0.6259

    def setup_config12(self):
        self.endtime = 0.25

        self.p1 = 0.4
        self.rho1 = 0.5313
        self.u1 = 0.0
        self.v1 = 0.0

        self.p2 = 1
        self.rho2 = 1.0
        self.u2 = 0.7276
        self.v2 = 0.0

        self.p3 = 1
        self.rho3 = 0.8
        self.u3 = 0.0
        self.v3 = 0.0

        self.p4 = 1.0
        self.rho4 = 1.0
        self.u4 = 0.0
        self.v4 = 0.7276


if __name__ == "__main__":
    config = R2DConfig()



"""Dam break flow against a tall structure

The case is from "Hakan Akyildiz, N. Erdem Unal,
Sloshing in a three-dimensional rectangular tank:
Numerical simulation and experimental validation,
Ocean Engineering, Volume 33, Issue 16,
2006, Pages 2135-2149, ISSN 0029-8018"
DOI: https://doi.org/10.1016/j.oceaneng.2005.11.001.

pitch angle = 4 deg, roll frequency = 2 rad/s, 75% filled

"""
from math import sin, pi, cos

import os
import numpy as np

# PySPH imports
from pysph.base.utils import get_particle_array
from pysph.base.kernels import CubicSpline
from pysph.solver.solver import Solver
from pysph.solver.application import Application

from pysph.sph.integrator_step import WCSPHStep, IntegratorStep
from pysph.sph.integrator import PECIntegrator

from pysph.sph.equation import Group, Equation

from pysph.sph.scheme import WCSPHScheme
from pysph.examples._db_geometry import DamBreak3DGeometry

Umax = np.sqrt(9.81*0.75*0.62)
c0 = 10.0 * Umax

dx = 0.02
hdx = 1.2

h0 = hdx * dx

tf = 10.0
rho = 1000.0

alpha = 0.1
beta = 0.0
gamma = 7.0

length = 0.92
width = 0.46
height = 0.62
n_layers = 3

theta_0 = 4 * pi/180
omega_r = 2


class PitchingMotion(Equation):
    def __init__(self, dest, sources, theta_0, omega_r):
        self.theta_0 = theta_0
        self.omega_r = omega_r
        super(PitchingMotion, self).__init__(dest, sources)

    def initialize(self, d_idx, d_au, d_aw, t, d_z, d_x):
        theta_0 = self.theta_0
        omega_r = self.omega_r

        omega = theta_0*omega_r*cos(omega_r*t)
        alpha = -theta_0*omega_r*omega_r*sin(omega_r*t)

        at_x = d_z[d_idx]*alpha
        at_z = -d_x[d_idx]*alpha

        ac_x = -d_x[d_idx]*omega*omega
        ac_z = -d_z[d_idx]*omega*omega

        d_au[d_idx] = at_x + ac_x
        d_aw[d_idx] = at_z + ac_z


class OneStageRigidBodyStep(IntegratorStep):
    def initialize(self, d_idx, d_x, d_y, d_z, d_x0, d_y0, d_z0,
                   d_u, d_v, d_w, d_u0, d_v0, d_w0, d_rho, d_rho0):

        d_u0[d_idx] = d_u[d_idx]
        d_v0[d_idx] = d_v[d_idx]
        d_w0[d_idx] = d_w[d_idx]

        d_x0[d_idx] = d_x[d_idx]
        d_y0[d_idx] = d_y[d_idx]
        d_z0[d_idx] = d_z[d_idx]

        d_rho0[d_idx] = d_rho[d_idx]

    def stage1(self, d_idx, d_x, d_y, d_z, d_x0, d_y0, d_z0,
               d_u, d_v, d_w, d_u0, d_v0, d_w0, d_au, d_av, d_aw,
               dt, d_rho0, d_arho, d_rho):

        d_rho[d_idx] = d_rho0[d_idx] + dt * 0.5 * d_arho[d_idx]

    def stage2(self, d_idx, d_x, d_y, d_z, d_x0, d_y0, d_z0,
               d_u, d_v, d_w, d_u0, d_v0, d_w0, d_au, d_av, d_aw,
               dt, d_rho0, d_arho, d_rho):

        # update velocities
        d_u[d_idx] += dt * d_au[d_idx]
        d_v[d_idx] += dt * d_av[d_idx]
        d_w[d_idx] += dt * d_aw[d_idx]

        # update positions using time-centered velocity
        d_x[d_idx] += dt * 0.5 * (d_u[d_idx] + d_u0[d_idx])
        d_y[d_idx] += dt * 0.5 * (d_v[d_idx] + d_v0[d_idx])
        d_z[d_idx] += dt * 0.5 * (d_w[d_idx] + d_w0[d_idx])

        d_rho[d_idx] = d_rho0[d_idx] + dt * d_arho[d_idx]


class SloshingTankPitch(Application):
    def add_user_options(self, group):
        interp_methods = ['shepard', 'sph', 'order1']
        group.add_argument(
            '--dx', action='store', type=float, dest='dx', default=dx,
            help='Particle spacing.'
        )
        group.add_argument(
            '--hdx', action='store', type=float, dest='hdx', default=hdx,
            help='Specify the hdx factor where h = hdx * dx.'
        )
        group.add_argument(
            '--interp-method', action="store", type=str, dest='interp_method',
            default='shepard', help="Specify the interpolation method.",
            choices=interp_methods
        )

    def consume_user_options(self):
        self.hdx = self.options.hdx
        self.dx = self.options.dx
        self.h0 = self.hdx * self.dx
        self.interp_method = self.options.interp_method

    def create_particles(self):
        geom = DamBreak3DGeometry(
            container_height=height, container_width=width,
            container_length=length, fluid_column_height=height*0.75,
            fluid_column_width=width, fluid_column_length=length,
            nboundary_layers=n_layers, with_obstacle=False,
            dx=self.dx, hdx=self.hdx, rho0=rho
        )

        [fluid, boundary] = geom.create_particles()
        fluid.x = fluid.x - length*0.5
        boundary.x = boundary.x - length*0.5

        # Setting up intital velocity of the tank
        omega0 = theta_0 * omega_r
        boundary.u, boundary.w = boundary.z*omega0, -boundary.x*omega0

        self.scheme.setup_properties([fluid, boundary])
        particles = [fluid, boundary]
        return particles

    def create_solver(self):
        kernel = CubicSpline(dim=3)
        integrator = PECIntegrator(fluid=WCSPHStep(),
                                   boundary=OneStageRigidBodyStep())

        dt = 0.125 * self.h0 / c0
        self.scheme.configure(h0=self.h0, hdx=self.hdx)

        solver = Solver(kernel=kernel, dim=3, integrator=integrator,
                        tf=tf, dt=dt, adaptive_timestep=True,
                        fixed_h=False)

        return solver

    def create_scheme(self):
        s = WCSPHScheme(
            ['fluid'], ['boundary'], dim=3, rho0=rho, c0=c0,
            h0=h0, hdx=hdx, gz=-9.81, alpha=alpha,
            beta=beta, gamma=gamma, hg_correction=True,
            tensile_correction=False, delta_sph=True
        )
        return s

    def create_equations(self):
        eqns = self.scheme.get_equations()

        equation_1 = Group(
            equations=[
                PitchingMotion(
                    dest='boundary', sources=None,
                    theta_0=theta_0, omega_r=omega_r),
            ], real=False
        )

        eqns.insert(0, equation_1)

        return eqns

    def create_tools(self):
        tools = []
        from pysph.solver.tools import DensityCorrection
        rho_corr = DensityCorrection(self, ['fluid', 'boundary'],
                                     corr="shepard", freq=10,
                                     kernel=CubicSpline)

        tools.append(rho_corr)
        return tools

    def post_process(self, info_fname):
        self.read_info(info_fname)
        if len(self.output_files) == 0:
            return

        from pysph.solver.utils import iter_output
        import matplotlib.pyplot as plt
        from pysph.examples import st_exp_data as st
        from pysph.tools.interpolator import Interpolator

        files = self.output_files

        t = []
        p0 = []
        xc0 = 0.0
        for sd, arrays1, arrays2 in iter_output(files, "fluid", "boundary"):
            t.append(sd["t"])
            xc = arrays2.x.mean()
            if sd["t"] == 0:
                xc0 = xc
            # Point 1: Bottom right corner of the tank
            # Point 2: Top right corner of the tank
            # Point 3: Bottom left corner of the tank
            if xc > xc0:
                z1 = arrays2.z.min()
                x2 = arrays2.x.max()
                x1 = arrays2.x[np.where(arrays2.z == z1)[0][0]]
                z2 = arrays2.x[np.where(arrays2.x == x2)[0][0]]
                angle = np.arctan((z2-z1)/(x2-x1))
                x3 = arrays2.x.min()
                z3 = arrays2.z[np.where(arrays2.x == x3)[0][0]]

            if xc <= xc0:
                z2 = arrays2.z.max()
                x1 = arrays2.x.max()
                x2 = arrays2.x[np.where(arrays2.z == z2)[0][0]]
                z1 = arrays2.z[np.where(arrays2.x == x1)[0][0]]
                angle = np.arctan((z2-z1)/(x2-x1)) + pi
                z3 = arrays2.z.min()
                x3 = arrays2.x[np.where(arrays2.z == z3)[0][0]]

            if x3-x1 == 0:
                vec = np.array([-1, 0])
            else:
                vec = np.array([x3-x1, z3-z1])
                vec = vec/np.linalg.norm(vec)

            dx = self.dx
            # Probes at 0.17 m above Point 1 and 0.45 m above Point 3
            x_probe = [x1 + 0.17*cos(angle) + 3.25*dx*vec[0], x3 + 0.45*cos(angle) - 3.25*dx*vec[0]]
            z_probe = [z1 + 0.17*sin(angle) + 3.25*dx*vec[1], z3 + 0.45*sin(angle) - 3.25*dx*vec[1]]
            y_probe = [0, 0]
            interp = Interpolator([arrays1, arrays2], x=x_probe,
                                  y=y_probe, z=z_probe,
                                  method=self.interp_method)
            p0.append(interp.interpolate('p'))

        p0 = np.array(p0)
        p2 = p0.T[0]/1000
        p8 = p0.T[1]/1000

        p2_avg, p8_avg = p2, p8

        fname = os.path.join(self.output_dir, 'results.npz')
        t, p0 = list(map(np.asarray, (t, p0)))
        np.savez(fname, t=t, p0=p0)

        exp_t2, exp_p2, exp_t8, exp_p8 = st.get_au_pitch_data()

        figure_1 = plt.figure()
        # Probe 2
        plt.plot(t[:-10], p2_avg[:-10], label="Computed", figure=figure_1)
        plt.scatter(exp_t2, exp_p2, color=(0, 0, 0),
                    label="Experiment (Akyildiz and Unal, 2006)",
                    figure=figure_1)
        plt.title("P2 v/s t")
        plt.legend()
        plt.ylabel("Pressue [KPa]")
        plt.xlabel("Time [s]")
        plt.xlim(3, 10)
        plt.savefig(os.path.join(self.output_dir, 'p2_vs_t.png'))
        plt.show()

        figure_2 = plt.figure()
        # Probe 8
        plt.plot(t, p8_avg, label="Computed", figure=figure_2)
        plt.scatter(exp_t8, exp_p8, color=(0, 0, 0),
                    label="Experiment (Akyildiz and Unal, 2006)",
                    figure=figure_2)
        plt.title("P8 v/s t")
        plt.legend()
        plt.ylabel("Pressue [KPa]")
        plt.xlabel("Time [s]")
        plt.xlim(3, 10)
        plt.savefig(os.path.join(self.output_dir, 'p8_vs_t.png'))
        plt.show()


if __name__ == '__main__':
    app = SloshingTankPitch()
    app.run()
    app.post_process(app.info_filename)



from collections import defaultdict
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

from compyle.config import get_config
from pysph.sph.equation import (
    CUDAGroup, CythonGroup, Group, MultiStageEquations, OpenCLGroup,
    get_arrays_used_in_equation)


###############################################################################
def group_equations(equations):
    """Checks the given equations and ensures the following:

     - Raises an error if the user  mixes Groups and Equations.

     - If only equations are given as a list, return a single group with all
       these equations.
    """
    only_groups = [x for x in equations if isinstance(x, Group)]
    if len(only_groups) > 0 and len(only_groups) != len(equations):
        raise ValueError('All elements must be Groups if you use groups.')
    if len(only_groups) == 0:
        return [Group(equations)]
    else:
        return equations


###############################################################################
def check_equation_array_properties(equation, particle_arrays):
    """Given an equation and the particle arrays, check if the particle arrays
    have the necessary properties.
    """
    p_arrays = dict((x.name, x) for x in particle_arrays)
    _src, _dest = get_arrays_used_in_equation(equation)
    if equation.dest not in p_arrays:
        msg = "ERROR: Equation {eq_name} has invalid dest: '{dest}'".format(
            eq_name=equation.name, dest=equation.dest
        )
        raise RuntimeError(msg)
    if not equation.no_source:
        for src in equation.sources:
            if src not in p_arrays:
                msg = "ERROR: Equation {eq_name} has invalid "\
                      "source: '{src}'".format(eq_name=equation.name, src=src)
                raise RuntimeError(msg)

    eq_src = set([x[2:] for x in _src])
    eq_dest = set([x[2:] for x in _dest])

    def _check_array(array, eq_props, errors):
        """Updates the `errors` with any errors.
        """
        props = set(list(array.properties.keys()) +
                    list(array.constants.keys()))
        if not eq_props < props:
            errors[array.name].update(eq_props - props)

    errors = defaultdict(set)
    _check_array(p_arrays[equation.dest], eq_dest, errors)
    if equation.sources is not None:
        for src in equation.sources:
            _check_array(p_arrays[src], eq_src, errors)

    if len(errors) > 0:
        msg = ("ERROR: Missing array properties for equation: %s\n"
               % equation.name)
        for name, missing in errors.items():
            msg += "Array '%s' missing properties %s.\n" % (name, missing)
        print(msg)
        raise RuntimeError(msg)


def make_acceleration_evals(particle_arrays, equations, kernel,
                            mode='serial', backend=None):
    '''Returns a list of acceleration evaluators.

    If a MultiStageEquations object is given the resulting list will have
    multiple evaluators else it will have a single one.
    '''
    if isinstance(equations, MultiStageEquations):
        groups = equations.groups
    else:
        groups = [equations]
    return [
        AccelerationEval(particle_arrays, group, kernel, mode, backend)
        for group in groups
    ]


###############################################################################
class MegaGroup(object):
    """A mega-group refactors actual equation Groups into a more
    organized form as described below.  They inherit all properties of the
    Group so these can be used while generating code but delegate the tasks to
    real groups underneath that are assembled from the original group.

    MegaGroups are organized as:
        {destination: (eqs_with_no_source, sources, all_eqs)}
        eqs_with_no_source: Group([equations]) all SPH Equations with no src.
        sources are {source: Group([equations...])}
        all_eqs is a Group of all equations having this destination.

    This is what is stored in the `data` attribute.

    Note that the order of the equations in all_eqs, eqs_with_no_source and
    sources should honor the order in which the user defines them in the
    original group.
    """
    def __init__(self, group, group_cls):
        self._orig_group = group
        self.Group = group_cls
        self._copy_props(group)
        self.data = self._make_data(group)

    def get_converged_condition(self):
        return self._orig_group.get_converged_condition()

    def _copy_props(self, group):
        for key in ('real', 'update_nnps', 'iterate', 'pre', 'post',
                    'max_iterations', 'min_iterations', 'has_subgroups',
                    'condition', 'start_idx', 'stop_idx'):
            setattr(self, key, getattr(group, key))

    def _make_data(self, group):
        equations = group.equations

        if group.has_subgroups:
            return [MegaGroup(g, self.Group) for g in equations]

        dest_list = []
        for equation in equations:
            dest = equation.dest
            if dest not in dest_list:
                dest_list.append(dest)

        dests = OrderedDict()
        for dest in dest_list:
            sources = defaultdict(list)
            eqs_with_no_source = []
            all_equations = []
            for equation in equations:
                if equation.dest != dest:
                    continue
                if equation not in all_equations:
                    all_equations.append(equation)
                if equation.no_source:
                    eqs_with_no_source.append(equation)
                else:
                    for src in equation.sources:
                        sources[src].append(equation)

            for src in sources:
                eqs = sources[src]
                sources[src] = self.Group(eqs)

            dests[dest] = (self.Group(eqs_with_no_source), sources,
                           self.Group(all_equations))

        return dests


###############################################################################
class AccelerationEval(object):
    def __init__(self, particle_arrays, equations, kernel, mode='serial',
                 backend=None):
        """

        Parameters
        ----------

        particle_arrays: list(ParticleArray): list of particle arrays to use.
        equations: list: A list of equations/groups.
        kernel: The kernel to use.
        mode: str: One of 'serial', 'mpi'.
        backend: str: indicates the backend to use.
            one of ('opencl', 'cython', 'cuda', '', None)
        """
        assert backend in ('opencl', 'cython', 'cuda', '', None)
        self.backend = self._get_backend(backend)
        self.particle_arrays = particle_arrays
        self.equation_groups = group_equations(equations)
        self.kernel = kernel
        self.nnps = None
        self.mode = mode
        if self.backend == 'cython':
            self.Group = CythonGroup
        elif self.backend == 'opencl':
            self.Group = OpenCLGroup
        elif self.backend == 'cuda':
            self.Group = CUDAGroup

        all_equations = []
        for group in self.equation_groups:
            if group.has_subgroups:
                for g in group.equations:
                    all_equations.extend(g.equations)
            else:
                all_equations.extend(group.equations)
        self.all_group = self.Group(equations=all_equations)

        for equation in all_equations:
            check_equation_array_properties(equation, particle_arrays)

        self.mega_groups = [MegaGroup(g, self.Group)
                            for g in self.equation_groups]
        self.c_acceleration_eval = None

    ##########################################################################
    # Private interface.
    ##########################################################################
    def _get_backend(self, backend):
        if not backend:
            cfg = get_config()
            if cfg.use_opencl:
                backend = 'opencl'
            elif cfg.use_cuda:
                backend = 'cuda'
            else:
                backend = 'cython'
        return backend

    ##########################################################################
    # Public interface.
    ##########################################################################
    def compute(self, t, dt):
        """Compute the accelerations given the current time, t, and the
        timestep, dt.
        """
        self.c_acceleration_eval.compute(t, dt)

    def set_compiled_object(self, c_acceleration_eval):
        """Set the high-performance compiled object to call internally.
        """
        self.c_acceleration_eval = c_acceleration_eval

    def set_nnps(self, nnps):
        self.nnps = nnps
        self.c_acceleration_eval.set_nnps(nnps)

    def update_particle_arrays(self, particle_arrays):
        """Call this to update the particle arrays with new ones.  Make sure
        though that the same properties exist in both or you will get a
        segfault.
        """
        self.c_acceleration_eval.update_particle_arrays(particle_arrays)
